from temboardui.application import (
    add_role,
    add_role_in_group,
    check_role_email,
    check_role_name,
    check_role_password,
    check_role_phone,
    delete_role,
    delete_role_from_group,
    get_group_list,
    get_groups_by_role,
    get_role,
    get_role_list,
    hash_password,
    update_role,
)
from temboardui.web import (
    HTTPError,
    admin_required,
    app,
    render_template,
)
from temboardui.errors import TemboardUIError


@app.route(r'/settings/users')
@admin_required
def users(request):
    role_list = get_role_list(request.db_session)
    return render_template(
        'settings/user.html',
        nav=True, role=request.current_user,
        role_list=role_list
    )


@app.route(r'/json/settings/user', methods=['POST'])
@admin_required
def create_user(request):
    data = request.json
    validate_user_data(data)

    h_passwd = handle_password(data)
    role = add_role(request.db_session,
                    data['new_username'],
                    h_passwd,
                    data['email'],
                    data['is_active'],
                    data['is_admin'])
    add_user_to_groups(request, data, role)
    return {'message': 'OK'}


@app.route(r"/json/settings/user/([0-9a-z\-_\.]{3,16})$",
           methods=['GET', 'POST'])
@admin_required
def json_user(request, username):
    if 'GET' == request.method:
        if username is None:
            raise HTTPError(500, "Username is missing")
        role = get_role(request.db_session, username)
        groups = get_group_list(request.db_session)

        return {
            'role_name': role.role_name,
            'role_email': role.role_email,
            'role_phone': role.role_phone,
            'is_active': role.is_active,
            'is_admin': role.is_admin,
            'in_groups': [group.group_name for group in role.groups],
            'groups': [{
                'name': group.group_name,
                'description': group.group_description
            } for group in groups]
        }
    elif 'POST' == request.method:  # update
        data = request.json
        role = get_role(request.db_session, username)
        validate_user_data(data, role)

        h_passwd = handle_password(data)

        # First step is to remove user from the groups he belongs to.
        role_groups = get_groups_by_role(request.db_session,
                                         role.role_name)
        if role_groups:
            for role_group in role_groups:
                delete_role_from_group(request.db_session, role.role_name,
                                       role_group.group_name)
        role = update_role(
            request.db_session,
            role.role_name,
            data['new_username'],
            h_passwd,
            data['email'],
            data['is_active'],
            data['is_admin'],
            data['phone'])

        add_user_to_groups(request, data, role)

        return {"message": "OK"}


@app.route(r'/json/settings/delete/user', methods=['POST'])
@admin_required
def delete_user(request):
    data = request.json
    if not data.get('username'):
        raise HTTPError(400, "Username field is missing.")
    delete_role(request.db_session, data['username'])
    return {'delete': True}


def validate_user_data(data, role=None):
    # Submited attributes checking.
    if not data.get('new_username'):
        raise TemboardUIError(400, "Username is missing.")
    if data.get('email'):
        check_role_email(data['email'])
    if data.get('phone'):
        check_role_phone(data['phone'])
    if 'groups' not in data:
        raise TemboardUIError(400, "Groups field is missing.")
    if 'is_active' not in data:
        raise TemboardUIError(400, "Active field is missing.")
    if 'is_admin' not in data:
        raise TemboardUIError(400, "Administrator field is missing.")

    if role and role.role_name != data['new_username']:
        if not data.get('password'):
            raise TemboardUIError(
                400, "Username will be changed, you need to change "
                "the password too.")
    if role is None:
        if not data.get('password'):
            raise TemboardUIError(400, "Password is missing.")
    if data.get('password') and not data.get('password2'):
        raise TemboardUIError(400, "Password confirmation is missing.")
    if 'password' in data and 'password2' in data:
        if data['password'] != data['password2']:
            raise TemboardUIError(
                400, "Password confirmation can not be checked.")
    if data['groups'] is not None and type(data['groups']) != list:
        raise TemboardUIError(400, "Invalid group list.")

    check_role_name(data['new_username'])


def handle_password(data):
    if data['password']:
        check_role_password(data['password'])
        return hash_password(data['new_username'], data['password'])
    return None


def add_user_to_groups(request, data, role):
    if data['groups']:
        for group_name in data['groups']:
            add_role_in_group(request.db_session, role.role_name, group_name)



from temboardui.application import (
    get_instance_groups_by_role,
    get_instances_by_role_name,
)
from ..web import (
    Redirect,
    app,
    anonymous_allowed,
    jsonify,
    render_template,
)
from ..plugins.monitoring.alerting import get_highest_state


@app.route('/')
@anonymous_allowed
def index(request):
    return Redirect('/home')


@app.route('/home')
def home(request):
    role = request.current_user

    groups = get_instance_groups_by_role(request.db_session, role.role_name)
    groups = [group for group in groups]

    return render_template(
        'home.html',
        nav=True, role=role,
        groups=groups,
    )


@app.route('/home/instances')
def home_instances(request):
    role = request.current_user

    # get the instances for which current user has access to
    instances = get_instances_by_role_name(request.db_session, role.role_name)
    instances = [{
        'hostname': instance.hostname,
        'agent_address': instance.agent_address,
        'agent_port': instance.agent_port,
        'pg_data': instance.pg_data,
        'pg_port': instance.pg_port,
        'pg_version': instance.pg_version,
        'pg_version_summary': instance.pg_version_summary,
        'groups': [group.group_name for group in instance.groups],
        'plugins': [plugin.plugin_name for plugin in instance.plugins],
    } for instance in instances]

    # Get availability for all monitored instances
    cur = request.db_session.connection().connection.cursor()
    cur.execute("SET search_path TO monitoring")
    sql = """
        SELECT
          distinct(ia.instance_id),
          i.port,
          h.hostname,
          first_value(available) OVER
            (PARTITION BY ia.instance_id ORDER BY datetime desc) AS available
        FROM monitoring.instance_availability AS ia
        JOIN monitoring.instances AS i
        ON i.instance_id = ia.instance_id
        JOIN monitoring.hosts AS h
        ON i.host_id = h.host_id;
    """
    all_availability = request.db_session.execute(sql, {}).fetchall()
    all_availability = {(i.hostname, i.port): i.available
                        for i in all_availability}

    # set instances availability if known
    for instance in instances:
        try:
            available = all_availability.get((instance['hostname'],
                                              instance['pg_port']))
        except KeyError:
            # Instance may not be monitored yet
            # because it has been added recently or monitor plugin is not
            # activated
            available = None
        instance['available'] = available

    sql = """
        WITH states_by_key AS (
            SELECT
                check_id,
                json_agg(json_build_object('key', key, 'state', state))
                  as state_by_key
            FROM monitoring.check_states
            GROUP BY check_id
        )
        SELECT h.hostname, i.port,
        json_agg(
            json_build_object(
                'name', c.name,
                'warning', c.warning,
                'critical', c.critical,
                'description', c.description,
                'enabled', c.enabled,
                'state_by_key', states_by_key.state_by_key
        )) AS checks
        FROM monitoring.checks c
        JOIN states_by_key ON c.check_id = states_by_key.check_id
        JOIN monitoring.instances i
          ON c.host_id = i.host_id AND c.instance_id = i.instance_id
        JOIN monitoring.hosts h ON h.host_id = i.host_id
        GROUP BY 1,2;
    """
    all_checks = request.db_session.execute(sql, {}).fetchall()
    all_checks = {
        (i.hostname, i.port): [{
                'name': check['name'],
                'state': get_highest_state([s['state']
                                            for s in check['state_by_key']]),
                'description': check['description']
            } for check in i.checks]
        for i in all_checks
    }

    # set instances checks if any
    for instance in instances:
        try:
            checks = all_checks[(instance['hostname'], instance['pg_port'])]
        except KeyError:
            # Instance may not be monitored yet
            # because it has been added recently or monitor plugin is not
            # activated
            checks = []
        instance['checks'] = checks

    return jsonify(instances)



import logging

from temboardui.web import (
    HTTPError,
    admin_required,
    app,
    render_template,
)

from temboardui.application import (
    get_group_list,
    get_group,
    check_group_name,
    check_group_description,
    delete_role_group_from_instance_group,
    update_group,
    add_group,
    add_role_group_in_instance_group,
    delete_group,
)


logger = logging.getLogger(__name__)
PREFIX = r'/json/settings'


@app.route(PREFIX + r'/all/group/(role|instance)$')
@admin_required
def all_group(request, kind):
    groups = get_group_list(request.db_session, kind)
    return {
        'groups': [{
            'name': group.group_name,
            'kind': group.group_kind,
            'description': group.group_description
        } for group in groups],
        'loaded_plugins': request.handler.application.loaded_plugins,
    }


@app.route(
    PREFIX + r"/group/(role|instance)(?:/([0-9a-z\-_\.]{3,16}))?$",
    methods=['GET', 'POST'])
@admin_required
def group(request, kind, name):
    if 'GET' == request.method:
        if not name:
            raise HTTPError(404)
        group = get_group(request.db_session, name, kind)
        data = dict(
            name=group.group_name,
            kind=kind,
            description=group.group_description,
        )
        if kind == 'instance':
            data['user_groups'] = [
                dict(name=g.group_name, description=g.group_description)
                for g in get_group_list(request.db_session)
            ]
            data['in_groups'] = [a.role_group_name for a in group.ari]
        return data
    else:  # POST
        data = request.json
        if not data.get('new_group_name'):
            raise HTTPError(400, "Missing group name.")
        check_group_name(data['new_group_name'])
        if not data.get('description'):
            raise HTTPError(400, "Missing description")
        check_group_description(data['description'])

        if name:  # Group update
            group = get_group(request.db_session, name, kind)
            if 'instance' == kind:
                for ari in group.ari:
                    delete_role_group_from_instance_group(
                        request.db_session, ari.role_group_name,
                        group.group_name)

            group = update_group(
                request.db_session,
                group.group_name, kind,
                data['new_group_name'], data['description'])
        else:  # Group creation
            group = add_group(
                request.db_session,
                data['new_group_name'], data['description'],
                kind)

        if 'user_groups' in data and data['user_groups']:
            for group_name in data['user_groups']:
                add_role_group_in_instance_group(
                    request.db_session, group_name, group.group_name)

        return {'ok': True}


@app.route(PREFIX + r"/delete/group/(role|instance)", methods=['POST'])
@admin_required
def delete_group_handler(request, kind):
    name = request.json.get('group_name')
    if not name:
        raise HTTPError(400)
    delete_group(request.db_session, name, kind)
    return {'delete': True}


@app.route(r"/settings/groups/(role|instance)")
@admin_required
def groups(request, kind):
    return render_template(
        "settings/group.html",
        nav=True, role=request.current_user,
        group_kind=kind,
        group_list=get_group_list(request.db_session, kind),
    )



# coding: utf-8
from textwrap import dedent


def insert_availability(session, dt, instance_id, available):
    session.execute(
        dedent("""
        SELECT monitoring.insert_instance_availability(
            :dt,
            :instance_id,
            :available
        )
        """),
        dict(dt=dt, instance_id=instance_id, available=available)
    )


def insert_metric_sessions(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_sessions_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                metric['active'],
                metric['waiting'],
                metric['idle'],
                metric['idle_in_xact'],
                metric['idle_in_xact_aborted'],
                metric['fastpath'],
                metric['disabled'],
                metric['no_priv']
            )
        )
    )


def insert_metric_xacts(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_xacts_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                str(metric['measure_interval']),
                metric['n_commit'],
                metric['n_rollback']
            )
        )
    )


def insert_metric_locks(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_locks_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                metric['access_share'],
                metric['row_share'],
                metric['row_exclusive'],
                metric['share_update_exclusive'],
                metric['share'],
                metric['share_row_exclusive'],
                metric['exclusive'],
                metric['access_exclusive'],
                metric['siread'],
                metric['waiting_access_share'],
                metric['waiting_row_share'],
                metric['waiting_row_exclusive'],
                metric['waiting_share_update_exclusive'],
                metric['waiting_share'],
                metric['waiting_share_row_exclusive'],
                metric['waiting_exclusive'],
                metric['waiting_access_exclusive']
            )
        )
    )


def insert_metric_blocks(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_blocks_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                str(metric['measure_interval']),
                metric['blks_read'],
                metric['blks_hit'],
                metric['hitmiss_ratio']
            )
        )
    )


def insert_metric_bgwriter(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_bgwriter_current
            VALUES (:dt, :instance_id, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            row=(
                None,
                str(metric['measure_interval']),
                metric['checkpoints_timed'],
                metric['checkpoints_req'],
                metric['checkpoint_write_time'],
                metric['checkpoint_sync_time'],
                metric['buffers_checkpoint'],
                metric['buffers_clean'],
                metric['maxwritten_clean'],
                metric['buffers_backend'],
                metric['buffers_backend_fsync'],
                metric['buffers_alloc'],
                metric['stats_reset']
            )
        )
    )


def insert_metric_db_size(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_db_size_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                metric['size']
            )
        )
    )


def insert_metric_tblspc_size(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_tblspc_size_current
            VALUES (:dt, :instance_id, :spcname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            spcname=metric['spcname'],
            row=(
                None,
                metric['size']
            )
        )
    )


def insert_metric_fs_size(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_filesystems_size_current
            VALUES (:dt, :host_id, :mount_point, :row)
        """),
        dict(
            dt=metric['datetime'],
            host_id=host_id,
            mount_point=metric['mount_point'],
            row=(
                None,
                metric['used'],
                metric['total'],
                metric['device']
            )
        )
    )


def insert_metric_temp_files_size_delta(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_temp_files_size_delta_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                str(metric['measure_interval']),
                metric['size']
            )
        )
    )


def insert_metric_wal_files(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_wal_files_current
            VALUES (:dt, :instance_id, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            row=(
                None,
                str(metric['measure_interval']),
                metric['written_size'],
                metric['current_location'],
                metric['total'],
                metric['archive_ready'],
                metric['total_size']
            )
        )
    )


def insert_metric_cpu(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_cpu_current
            VALUES (:dt, :host_id, :cpu, :row)
        """),
        dict(
            dt=metric['datetime'],
            host_id=host_id,
            cpu=metric['cpu'],
            row=(
                None,
                str(metric['measure_interval']),
                metric['time_user'],
                metric['time_system'],
                metric['time_idle'],
                metric['time_iowait'],
                metric['time_steal']
            )
        )
    )


def insert_metric_process(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_process_current
            VALUES (:dt, :host_id, :row)
        """),
        dict(
            dt=metric['datetime'],
            host_id=host_id,
            row=(
                None,
                str(metric['measure_interval']),
                metric['context_switches'],
                metric['forks'],
                metric['procs_running'],
                metric['procs_blocked'],
                metric['procs_total']
            )
        )
    )


def insert_metric_memory(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_memory_current
            VALUES (:dt, :host_id, :row)
        """),
        dict(
            dt=metric['datetime'],
            host_id=host_id,
            row=(
                None,
                metric['mem_total'],
                metric['mem_used'],
                metric['mem_free'],
                metric['mem_buffers'],
                metric['mem_cached'],
                metric['swap_total'],
                metric['swap_used']
            )
        )
    )


def insert_metric_loadavg(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_loadavg_current
            VALUES (:dt, :host_id, :row)
        """),
        dict(
            dt=metric['datetime'],
            host_id=host_id,
            row=(
                None,
                metric['load1'],
                metric['load5'],
                metric['load15']
            )
        )
    )


def insert_metric_vacuum_analyze(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_vacuum_analyze_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                str(metric['measure_interval']),
                metric['n_vacuum'],
                metric['n_analyze'],
                metric['n_autovacuum'],
                metric['n_autoanalyze']
            )
        )
    )


def insert_metric_replication_lag(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_replication_lag_current
            VALUES (:dt, :instance_id, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            row=(
                None,
                metric['lag']
            )
        )
    )


def insert_metric_replication_connection(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_replication_connection_current
            VALUES (:dt, :instance_id, :upstream, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            upstream=metric['upstream'],
            row=(
                None,
                metric['connected']
            )
        )
    )


def insert_metric_heap_bloat(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_heap_bloat_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                metric['ratio']
            )
        )
    )


def insert_metric_btree_bloat(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_btree_bloat_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric['datetime'],
            instance_id=instance_id,
            dbname=metric['dbname'],
            row=(
                None,
                metric['ratio']
            )
        )
    )


def get_host_id(session, hostname):
    row = session.execute(
        "SELECT host_id FROM monitoring.hosts WHERE hostname = :hostname",
        dict(hostname=hostname)
    ).fetchone()
    return row[0] if row else None


def get_instance_id(session, host_id, port):
    row = session.execute(
        dedent("""
            SELECT instance_id
            FROM monitoring.instances
            WHERE host_id = :host_id AND port = :port
        """),
        dict(host_id=host_id, port=port)
    ).fetchone()
    return row[0] if row else None


def get_agent_key(session, hostname, pg_data, pg_port):
    row = session.execute(
        dedent("""
            SELECT agent_key
            FROM application.instances
            WHERE hostname = :hostname AND pg_data=:pg_data
            AND pg_port = :pg_port
            LIMIT 1
        """),
        dict(
            hostname=hostname,
            pg_data=pg_data,
            pg_port=pg_port
        )
    ).fetchone()
    return row[0] if row else None


def get_agent_keys(session, hostname):
    return session.execute(
        dedent("""
            SELECT agent_key
            FROM application.instances
            WHERE hostname = :hostname
        """),
        dict(
            hostname=hostname
        )
    ).fetchall()


def append_state_changes(
    session, dt, check_id, state, key, value, warning, critical
):
    session.execute(
        dedent("""
            SELECT monitoring.append_state_changes(
                :datetime, :check_id, :state, :key, :value, :warning, :critical
            )
        """),
        dict(
            datetime=dt,
            check_id=check_id,
            state=state,
            key=key,
            value=value,
            warning=warning,
            critical=critical
        )
    )


def purge_check_states(session, check_id, keys):
    session.execute(
        dedent("""
            DELETE FROM monitoring.check_states
            WHERE check_id = :check_id AND NOT (key = ANY(:keys))
        """),
        dict(
            check_id=check_id,
            keys=keys
        )
    )


def undef_check_states(session, all_ids, to_keep_ids):
    session.execute(
        dedent("""
            UPDATE monitoring.check_states
            SET state = 'UNDEF'
            WHERE
                check_id = ANY(:all_ids)
                AND NOT (check_id = ANY(:to_keep_ids))
        """),
        dict(
            all_ids=all_ids,
            to_keep_ids=to_keep_ids,
        )
    )



import cStringIO
import datetime
from psycopg2.extensions import AsIs

from .pivot import pivot_timeserie


METRICS = dict(
    blocks=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    ROUND(SUM((record).blks_read)/(extract('epoch' from MIN((record).measure_interval)))) AS blks_read_s,
    ROUND(SUM((record).blks_hit)/(extract('epoch' from MIN((record).measure_interval)))) AS blks_hit_s
FROM expand_data_by_instance_id('metric_blocks', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_blocks_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    ROUND(SUM((record).blks_read)/(extract('epoch' from MIN((record).measure_interval)))) AS blks_read_s,
    ROUND(SUM((record).blks_hit)/(extract('epoch' from MIN((record).measure_interval)))) AS blks_hit_s
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC
        """,  # noqa
        probename='blocks',
    ),
    checkpoints=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).checkpoints_timed AS timed,
    (record).checkpoints_req AS req,
    ROUND(((record).checkpoint_write_time/1000)::numeric, 1) AS write_time,
    ROUND(((record).checkpoint_sync_time/1000)::numeric,1) AS sync_time
FROM expand_data_by_instance_id('metric_bgwriter', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_bgwriter_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).checkpoints_timed AS timed,
    (record).checkpoints_req AS req,
    ROUND(((record).checkpoint_write_time/1000)::numeric, 1) AS write_time,
    ROUND(((record).checkpoint_sync_time/1000)::numeric,1) AS sync_time
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='bgwriter',
    ),
    cpu=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round((SUM((record).time_user)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS user,
    round((SUM((record).time_system)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS system,
    round((SUM((record).time_iowait)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS iowait,
    round((SUM((record).time_steal)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS steal
FROM expand_data_by_host_id('metric_cpu', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, cpu text, record metric_cpu_record)
GROUP BY datetime, host_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round((SUM((record).time_user)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS user,
    round((SUM((record).time_system)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS system,
    round((SUM((record).time_iowait)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS iowait,
    round((SUM((record).time_steal)/(SUM((record).time_user)+SUM((record).time_system)+SUM((record).time_idle)+SUM((record).time_iowait)+SUM((record).time_steal))::float*100)::numeric, 1) AS steal
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, host_id ORDER BY datetime
        """,  # noqa
        probename='cpu',
    ),
    cpu_core=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round(((record).time_user/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS user,
    round(((record).time_system/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS system,
    round(((record).time_iowait/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS iowait,
    round(((record).time_steal/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS steal
FROM expand_data_by_host_id('metric_cpu', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, cpu text, record metric_cpu_record)
WHERE cpu = %(key)s
ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round(((record).time_user/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS user,
    round(((record).time_system/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS system,
    round(((record).time_iowait/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS iowait,
    round(((record).time_steal/((record).time_user+(record).time_system+(record).time_idle+(record).time_iowait+(record).time_steal)::float*100)::numeric, 1) AS steal
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND cpu = %(key)s
ORDER BY datetime
        """,  # noqa
        probename='cpu',
    ),
    ctxforks=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round(SUM((record).context_switches)/(extract('epoch' from MIN((record).measure_interval)))) AS context_switches_s,
    round(SUM((record).forks)/(extract('epoch' from MIN((record).measure_interval)))) AS forks_s
FROM expand_data_by_host_id('metric_process', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, record metric_process_record)
GROUP BY datetime ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round(SUM((record).context_switches)/(extract('epoch' from MIN((record).measure_interval)))) AS context_switches_s,
    round(SUM((record).forks)/(extract('epoch' from MIN((record).measure_interval)))) AS forks_s
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime ORDER BY datetime
        """,  # noqa,
        probename='process',
    ),
    db_size=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    dbname,
    (record).size
FROM expand_data_by_instance_id('metric_db_size', tstzrange(%(start)s, %(end)s), %(instance_id))
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_db_size_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    dbname,
    (record).size
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY datetime, dbname
        """,  # noqa
        probename='db_size',
        pivot=dict(
            index='date',
            key='dbname',
            value='size',
        )
    ),
    fs_size=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    mount_point,
    (record).used AS size
FROM expand_data_by_host_id('metric_filesystems_size', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, mount_point text, record metric_filesystems_size_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    mount_point,
    (record).used AS size
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='filesystems_size',
        pivot=dict(
            index='date',
            key='mount_point',
            value='size',
        ),
    ),
    fs_usage=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    mount_point,
    round((((record).used::FLOAT/(record).total::FLOAT)*100)::numeric, 1) AS usage
FROM expand_data_by_host_id('metric_filesystems_size', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, mount_point text, record metric_filesystems_size_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    mount_point,
    round((((record).used::FLOAT/(record).total::FLOAT)*100)::numeric, 1) AS usage
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa,
        probename='filesystems_size',
        pivot=dict(
            index='date',
            key='mount_point',
            value='usage',
        ),
    ),
    fs_usage_mountpoint=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round((((record).used::FLOAT/(record).total::FLOAT)*100)::numeric, 1) AS usage
FROM expand_data_by_host_id('metric_filesystems_size', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, mount_point text, record metric_filesystems_size_record)
WHERE mount_point = %(key)s
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round((((record).used::FLOAT/(record).total::FLOAT)*100)::numeric, 1) AS usage
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND mount_point = %(key)s
ORDER BY 1,2 ASC
        """,  # noqa,
        probename='filesystems_size',
    ),
    hitreadratio=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    CASE WHEN (SUM((record).blks_hit) + SUM((record).blks_read)) > 0
    THEN ROUND((SUM((record).blks_hit)::FLOAT/(SUM((record).blks_hit) + SUM((record).blks_read)::FLOAT) * 100)::numeric, 2)
    ELSE 100 END AS hit_read_ratio
FROM expand_data_by_instance_id('metric_blocks', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_blocks_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    CASE WHEN (SUM((record).blks_hit) + SUM((record).blks_read)) > 0
    THEN ROUND((SUM((record).blks_hit)::FLOAT/(SUM((record).blks_hit) + SUM((record).blks_read)::FLOAT) * 100)::numeric, 2)
    ELSE 100 END AS hit_read_ratio
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC
        """,  # noqa
        probename='blocks',
    ),
    hitreadratio_db=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    CASE WHEN ((record).blks_hit + (record).blks_read) > 0
    THEN ROUND((((record).blks_hit::FLOAT/((record).blks_hit + (record).blks_read)::FLOAT) * 100)::numeric, 2)
    ELSE 100 END AS hit_read_ratio
FROM expand_data_by_instance_id('metric_blocks', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_blocks_record)
WHERE dbname = %(key)s
ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    CASE WHEN ((record).blks_hit + (record).blks_read) > 0
    THEN ROUND((((record).blks_hit::FLOAT/((record).blks_hit + (record).blks_read)::FLOAT) * 100)::numeric, 2)
    ELSE 100 END AS hit_read_ratio
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND dbname = %(key)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='blocks',
    ),
    instance_size=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    SUM((record).size) AS size
FROM expand_data_by_instance_id('metric_db_size', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_db_size_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    SUM((record).size) AS size
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC
        """,  # noqa
        probename='db_size',
    ),
    loadavg=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).load1,
    (record).load5,
    (record).load15
FROM expand_data_by_host_id( 'metric_loadavg', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, record metric_loadavg_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).load1,
    (record).load5,
    (record).load15
FROM %(tablename)s WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY datetime ASC
        """,  # noqa
        probename='loadavg',
    ),
    load1=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).load1
FROM expand_data_by_host_id( 'metric_loadavg', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, record metric_loadavg_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).load1
FROM %(tablename)s WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY datetime ASC
        """,  # noqa
        probename='loadavg',
    ),
    locks=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    SUM((record).access_share) AS access_share,
    SUM((record).row_share) AS row_share,
    SUM((record).row_exclusive) AS row_exclusive,
    SUM((record).share_update_exclusive) AS share_update_exclusive,
    SUM((record).share) AS share,
    SUM((record).share_row_exclusive) AS share_row_exclusive,
    SUM((record).exclusive) AS exclusive,
    SUM((record).access_exclusive) AS access_exclusive,
    SUM((record).siread) AS siread
FROM expand_data_by_instance_id('metric_locks', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_locks_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    SUM((record).access_share) AS access_share,
    SUM((record).row_share) AS row_share,
    SUM((record).row_exclusive) AS row_exclusive,
    SUM((record).share_update_exclusive) AS share_update_exclusive,
    SUM((record).share) AS share,
    SUM((record).share_row_exclusive) AS share_row_exclusive,
    SUM((record).exclusive) AS exclusive,
    SUM((record).access_exclusive) AS access_exclusive,
    SUM((record).siread) AS siread
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC        """,  # noqa
        probename='locks',
    ),
    memory=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).mem_free AS free,
    (record).mem_cached AS cached,
    (record).mem_buffers AS buffers,
    ((record).mem_used - (record).mem_cached - (record).mem_buffers) AS other
FROM expand_data_by_host_id('metric_memory', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, record metric_memory_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).mem_free AS free,
    (record).mem_cached AS cached,
    (record).mem_buffers AS buffers,
    ((record).mem_used - (record).mem_cached - (record).mem_buffers) AS other
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY datetime
        """,  # noqa
        probename='memory',
    ),
    memory_usage=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round(((((record).mem_total - (record).mem_free - (record).mem_cached)::FLOAT/(record).mem_total::FLOAT)*100)::numeric, 1) AS usage
FROM expand_data_by_host_id('metric_memory', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, record metric_memory_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round(((((record).mem_total - (record).mem_free - (record).mem_cached)::FLOAT/(record).mem_total::FLOAT)*100)::numeric, 1) AS usage
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY datetime
        """,  # noqa
        probename='memory',
    ),
    rollback_db=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    SUM((record).n_rollback) AS rollback
FROM expand_data_by_instance_id('metric_xacts', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_xacts_record)
WHERE dbname = %(key)s
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    SUM((record).n_rollback) AS rollback
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND key = %(key)s
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        probename='xacts',
    ),
    sessions=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    SUM((record).active) AS active,
    SUM((record).waiting) AS waiting,
    SUM((record).idle) AS idle,
    SUM((record).idle_in_xact) AS idle_in_xact,
    SUM((record).idle_in_xact_aborted) AS idle_in_xact_aborted,
    SUM((record).fastpath) AS fastpath,
    SUM((record).disabled) AS disabled
FROM expand_data_by_instance_id('metric_sessions', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_sessions_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    SUM((record).active) AS active,
    SUM((record).waiting) AS waiting,
    SUM((record).idle) AS idle,
    SUM((record).idle_in_xact) AS idle_in_xact,
    SUM((record).idle_in_xact_aborted) AS idle_in_xact_aborted,
    SUM((record).fastpath) AS fastpath,
    SUM((record).disabled) AS disabled
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC
        """,  # noqa,
        probename='sessions',
    ),
    sessions_usage=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round(((SUM((record).active + (record).waiting + (record).idle + (record).idle_in_xact + (record).idle_in_xact_aborted + (record).fastpath + (record).disabled)::FLOAT/(SELECT setting FROM pg_settings WHERE name = 'max_connections')::FLOAT)*100)::numeric, 1) AS session_usage
FROM expand_data_by_instance_id('metric_sessions', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_sessions_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round(((SUM((record).active + (record).waiting + (record).idle + (record).idle_in_xact + (record).idle_in_xact_aborted + (record).fastpath + (record).disabled)::FLOAT/(SELECT setting FROM pg_settings WHERE name = 'max_connections')::FLOAT)*100)::numeric, 1) AS session_usage
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC
        """,  # noqa,
        probename='sessions',
    ),
    swap=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).swap_used AS used
FROM expand_data_by_host_id('metric_memory', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, record metric_memory_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).swap_used AS used
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY datetime
        """,  # noqa
        probename='memory',
    ),
    swap_usage=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round((((record).swap_used::FLOAT/(record).swap_total::FLOAT)*100)::numeric, 1) AS usage
FROM expand_data_by_host_id('metric_memory', tstzrange(%(start)s, %(end)s), %(host_id)s)
AS (datetime timestamp with time zone, host_id integer, record metric_memory_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round((((record).swap_used::FLOAT/(record).swap_total::FLOAT)*100)::numeric, 1) AS usage
FROM %(tablename)s
WHERE host_id = %(host_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY datetime
        """,  # noqa
        probename='memory',
    ),
    tblspc_size=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    spcname,
    (record).size
FROM expand_data_by_instance_id('metric_tblspc_size', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, spcname text, record metric_tblspc_size_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    spcname,
    (record).size
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa,
        probename='tblspc_size',
        pivot=dict(
            index='date',
            key='spcname',
            value='size',
        ),
    ),
    tps=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round(SUM((record).n_commit)/(extract('epoch' from MIN((record).measure_interval)))) AS commit,
    round(SUM((record).n_rollback)/(extract('epoch' from MIN((record).measure_interval)))) AS rollback
FROM expand_data_by_instance_id('metric_xacts', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_xacts_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round(SUM((record).n_commit)/(extract('epoch' from MIN((record).measure_interval)))) AS commit,
    round(SUM((record).n_rollback)/(extract('epoch' from MIN((record).measure_interval)))) AS rollback
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        probename='xacts',
    ),
    waiting_locks=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    SUM((record).waiting_access_share) AS access_share,
    SUM((record).waiting_row_share) AS row_share,
    SUM((record).waiting_row_exclusive) AS row_exclusive,
    SUM((record).waiting_share_update_exclusive) AS share_update_exclusive,
    SUM((record).waiting_share) AS share,
    SUM((record).waiting_share_row_exclusive) AS share_row_exclusive,
    SUM((record).waiting_exclusive) AS exclusive,
    SUM((record).waiting_access_exclusive) AS access_exclusive
FROM expand_data_by_instance_id('metric_locks', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_locks_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    SUM((record).waiting_access_share) AS access_share,
    SUM((record).waiting_row_share) AS row_share,
    SUM((record).waiting_row_exclusive) AS row_exclusive,
    SUM((record).waiting_share_update_exclusive) AS share_update_exclusive,
    SUM((record).waiting_share) AS share,
    SUM((record).waiting_share_row_exclusive) AS share_row_exclusive,
    SUM((record).waiting_exclusive) AS exclusive,
    SUM((record).waiting_access_exclusive) AS access_exclusive
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC
        """,  # noqa
        probename='locks'
    ),
    waiting_sessions_db=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).waiting
FROM expand_data_by_instance_id('metric_sessions', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_sessions_record)
WHERE dbname = %(key)s
ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).waiting
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND dbname = %(key)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='sessions'
    ),
    wal_files_size=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).written_size,
    (record).total_size
FROM expand_data_by_instance_id('metric_wal_files', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_wal_files_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).written_size,
    (record).total_size
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='wal_files',
    ),
    wal_files_archive=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).archive_ready
FROM expand_data_by_instance_id('metric_wal_files', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_wal_files_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).archive_ready
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='wal_files',
    ),
    wal_files_count=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).archive_ready,
    (record).total
FROM expand_data_by_instance_id('metric_wal_files', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_wal_files_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).archive_ready,
    (record).total
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='wal_files',
    ),
    wal_files_rate=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    round(SUM((record).written_size)/(extract('epoch' from MIN((record).measure_interval)))) AS written_size_s
FROM expand_data_by_instance_id('metric_wal_files', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_wal_files_record)
GROUP BY datetime, instance_id ORDER BY datetime
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    round(SUM((record).written_size)/(extract('epoch' from MIN((record).measure_interval)))) AS written_size_s
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
GROUP BY datetime, instance_id ORDER BY 1,2 ASC
        """,  # noqa
        probename='wal_files',
    ),
    wal_files_total=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).total
FROM expand_data_by_instance_id('metric_wal_files', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_wal_files_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).total
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='wal_files',
    ),
    w_buffers=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).buffers_checkpoint AS checkpoint,
    (record).buffers_clean AS clean,
    (record).buffers_backend AS backend
FROM expand_data_by_instance_id('metric_bgwriter', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_bgwriter_record)
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).buffers_checkpoint AS checkpoint,
    (record).buffers_clean AS clean,
    (record).buffers_backend AS backend
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1,2 ASC
        """,  # noqa
        probename='bgwriter',
    ),
    replication_lag=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).lag AS lag
FROM expand_data_by_instance_id('metric_replication_lag', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, record metric_replication_lag_record)
ORDER BY 1
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).lag AS lag
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s
ORDER BY 1
        """,  # noqa
        probename='replication_lag',
    ),
    replication_connection=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).connected AS connected
FROM expand_data_by_instance_id('metric_replication_connection', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, upstream text, record metric_replication_connection_record)
WHERE upstream = %(key)s
ORDER BY 1
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).connected AS connected
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND upstream = %(key)s
ORDER BY 1
        """,  # noqa
        probename='replication_connection',
    ),
    temp_files_size_delta=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).size
FROM expand_data_by_instance_id('metric_temp_files_size_delta', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_temp_files_size_delta_record)
WHERE dbname = %(key)s
ORDER BY 1
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).size
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND dbname = %(key)s
ORDER BY 1
        """,  # noqa
        probename='temp_files_size_delta',
    ),
    heap_bloat=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).ratio
FROM expand_data_by_instance_id('metric_heap_bloat', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_bloat_ratio_record)
WHERE dbname = %(key)s
ORDER BY 1
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    (record).ratio
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND dbname = %(key)s
ORDER BY 1
        """,  # noqa
        probename='heap_bloat',
    ),
    btree_bloat=dict(
        sql_nozoom="""
SELECT
    datetime AS date,
    (record).ratio
FROM expand_data_by_instance_id('metric_btree_bloat', tstzrange(%(start)s, %(end)s), %(instance_id)s)
AS (datetime timestamp with time zone, instance_id integer, dbname text, record metric_bloat_ratio_record)
WHERE dbname = %(key)s
ORDER BY 1
        """,  # noqa
        sql_zoom="""
SELECT
    datetime AS date,
    dbname,
    (record).ratio
FROM %(tablename)s
WHERE instance_id = %(instance_id)s AND datetime >= %(start)s AND datetime <= %(end)s AND dbname = %(key)s
ORDER BY 1
        """,  # noqa
        probename='btree_bloat',
    ),
)


def zoom_level(start, end):
    zoom = 0
    if end:
        d = end - start
    else:
        d = datetime.datetime.utcnow() - start

    if d.days > 1 and d.days <= 31:
        zoom = 1
    elif d.days > 31 and d.days <= 365:
        zoom = 2
    elif d.days > 365:
        zoom = 2
    return zoom


def get_tablename(probename, zoom):
    if zoom == 1:
        return 'metric_%s_30m_current' % (probename)
    elif zoom == 2:
        return 'metric_%s_6h_current' % (probename)
    else:
        return


def get_metric_data_csv(session, metric_name, start, end, host_id=None,
                        instance_id=None, key=None):
    if metric_name not in METRICS:
        raise IndexError("Metric '%s' not found" % metric_name)

    metric = METRICS.get(metric_name)
    # Instanciate a new string buffer needed by copy_expert()
    data_buffer = cStringIO.StringIO()
    # Get a new psycopg2 cursor from the current sqlalchemy session
    cur = session.connection().connection.cursor()
    # Change working schema to 'monitoring'
    cur.execute("SET search_path TO monitoring")
    # Get the "zoom level", depending on the time interval
    level = zoom_level(start, end)
    # Load query template
    q_tpl = metric.get('sql_nozoom') if level == 0 else metric.get('sql_zoom')
    tablename = get_tablename(metric.get('probename'), level)
    query = cur.mogrify(q_tpl, dict(host_id=host_id, instance_id=instance_id,
                                    start=start, end=end, key=key,
                                    tablename=AsIs(tablename)))
    # Retreive data using copy_expert()
    cur.copy_expert("COPY(" + query + ") TO STDOUT WITH CSV HEADER",
                    data_buffer)
    cur.close()

    if metric.get('pivot'):
        # Apply pivot rotation
        data_pivot = cStringIO.StringIO()
        pivot_timeserie(
            data_buffer,
            index=metric.get('pivot').get('index'),
            key=metric.get('pivot').get('key'),
            value=metric.get('pivot').get('value'),
            output=data_pivot
        )

        data = data_pivot.getvalue()
        data_buffer.close()
        data_pivot.close()
    else:
        data = data_buffer.getvalue()
        data_buffer.close()
    return data


def get_unavailability_csv(session, start, end, host_id, instance_id):

    # Tell when the instance was not available
    cur = session.connection().connection.cursor()
    cur.execute("SET search_path TO monitoring")
    sql = """
        SELECT datetime FROM instance_availability
        WHERE instance_id = %(instance_id)s
        AND available = 'f'
        AND datetime >= %(start)s AND datetime <= %(end)s
    """
    query = cur.mogrify(sql, dict(instance_id=instance_id,
                                  start=start, end=end))
    data_buffer = cStringIO.StringIO()
    cur.copy_expert("COPY(" + query + ") TO STDOUT", data_buffer)
    data = data_buffer.getvalue()
    data_buffer.close()

    return data


def get_availability(session, host_id, instance_id):

    # Tell whether the instance is currently available or not
    cur = session.connection().connection.cursor()
    cur.execute("SET search_path TO monitoring")
    sql = """
        SELECT available FROM instance_availability
        WHERE instance_id = :instance_id
        ORDER BY datetime desc
        LIMIT 1
    """
    result = session.execute(sql, {'instance_id': instance_id}).fetchone()
    return bool(result[0]) if result else False



from datetime import datetime
import logging

from temboardui.web import (
    HTTPError,
    anonymous_allowed,
    csvify,
)

from .. import PLUGIN_NAME
from . import blueprint, render_template
from ..chartdata import (
    get_unavailability_csv,
    get_metric_data_csv,
)
from ..tools import (
    build_check_task_options,
    check_agent_key,
    check_host_key,
    get_instance_checks,
    get_instance_id,
    get_request_ids,
    insert_metrics,
    merge_agent_info,
    parse_start_end,
    populate_host_checks,
    update_collector_status,
)
from ..model.db import insert_availability

logger = logging.getLogger(__name__)


def check_agent_request(request, hostname, instance):
    key = request.headers.get('X-Key')
    if not key:
        raise HTTPError(401, 'X-Key header missing')

    if instance['available']:
        check_agent_key(request.db_session,
                        hostname,
                        instance['data_directory'],
                        instance['port'],
                        key)
    else:
        # Case when PostgreSQL instance is not started.
        check_host_key(request.db_session, hostname, key)


@blueprint.route(r"/(?:monitoring|supervision)/collector",
                 methods=['POST'], json=True)
@anonymous_allowed
def collector(request):
    data = request.json

    try:
        # Ignore legacy multi-instance.
        instance = data['instances'][0]
        hostinfo = data['hostinfo']
        hostname = hostinfo['hostname']
        port = instance['port']
        metrics_data = data['data']
        n_cpu = hostinfo['cpu_count']
        if 'max_connections' in instance:
            data['data']['max_connections'] = instance['max_connections']
    except (KeyError, IndexError) as e:
        logger.exception(str(e))
        raise HTTPError(409, "Not valid data")

    check_agent_request(request, hostname, instance)

    # Update the inventory
    host = merge_agent_info(request.db_session, hostinfo, instance)

    # Send the write SQL commands to the database because the metrics are
    # inserted with queries not the orm. Tables must be there.
    request.db_session.commit()

    instance_id = get_instance_id(request.db_session, host.host_id, port)

    insert_availability(
        request.db_session,
        data['datetime'],
        instance_id,
        data['instances'][0]['available']
    )
    request.db_session.commit()

    insert_metrics(request.db_session, host.host_id, instance_id, metrics_data)
    request.db_session.commit()

    # Update collector status
    update_collector_status(
        request.db_session,
        instance_id,
        u'OK',
        last_push=datetime.utcnow(),
        # This is the datetime format used by the agent
        last_insert=datetime.strptime(
            data['datetime'], "%Y-%m-%d %H:%M:%S +0000"
        ),
    )
    request.db_session.commit()

    # ALERTING PART
    populate_host_checks(request.db_session, host.host_id, instance_id,
                         dict(n_cpu=n_cpu))
    request.db_session.commit()

    # Create new task for checking preprocessed values
    task_options = build_check_task_options(
        request.json['data'], host.host_id, instance_id,
        get_instance_checks(request.db_session, instance_id),
        request.json['datetime'],
    )
    request.handler.application.temboard_app.scheduler.schedule_task(
        'check_data_worker',
        options=task_options,
        expire=0,
    )

    return {'done': True}


@blueprint.instance_route("/monitoring")
def index(request):
    request.instance.check_active_plugin(PLUGIN_NAME)
    try:
        agent_username = request.instance.get_profile()['username']
    except Exception:
        # Monitoring plugin doesn't require agent authentication since we
        # already have the data.
        # Don't fail if there's a session error (for example when the agent
        # has been restarted)
        agent_username = None

    return render_template(
        'index.html',
        nav=True, role=request.current_user, instance=request.instance,
        plugin='monitoring', agent_username=agent_username,
    )


@blueprint.instance_route("/monitoring/unavailability")
def unavailability(request):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)
    data = get_unavailability_csv(
        request.db_session, start, end, host_id, instance_id)
    return csvify(data)


@blueprint.instance_route(r'/monitoring/data/([a-z\-_.0-9]{1,64})$')
def data_metric(request, metric_name):
    key = request.handler.get_argument('key', default=None)
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)
    try:
        data = get_metric_data_csv(
            request.db_session, metric_name,
            start, end,
            host_id=host_id,
            instance_id=instance_id,
            key=key,
        )
    except IndexError:
        raise HTTPError(404, 'Unknown metric.')

    return csvify(data=data)



# -*- coding: utf-8 -*-

import os
import sys

from setuptools import setup, find_packages, __version__ as setuptoolsv


# Load version number
__version__ = None
setup_path = os.path.dirname(os.path.realpath(__file__))
exec(open(os.path.join(setup_path, 'temboardui', 'version.py'), 'r').read())

if setuptoolsv < '1.0':
    __version__ = __version__.replace('+', '.')

# Accept Tornado 5.X on Python 2.7.9+
BLEEDING_EDGE_TORNADO = '4.5' if sys.version_info < (2, 7, 9) else '6'


SETUP_KWARGS = dict(
    name='temboard',
    version=__version__,  # noqa, imported by execfile.
    description='temBoard User Interface.',
    author='Julien Tachoires, Étienne BERSAC',
    license='PostgreSQL',
    install_requires=[
        'alembic',
        'futures',
        'python-dateutil>=1.5',
        # There is no hard dependency on psycopg2 to allow using
        # psycopg2-binary instead. psycopg2 is not provided by psycopg2-binary
        # and there is no way to state an OR dependency in Python. It's up to
        # the user or package manager to ensure psycopg2 dependency. See
        # documentation.
        'sqlalchemy>=0.9.8',
        'tornado>=3.2,<' + BLEEDING_EDGE_TORNADO,
    ],
    include_package_data=True,
    zip_safe=False,
    url='https://github.com/dalibo/temboard/',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: System :: Monitoring",
    ],
    data_files=[
        ('share/temboard', [
            'share/auto_configure.sh',
            'share/create_repository.sh',
            'share/purge.sh',
        ]),
        ('share/temboard/sql/', [
            'share/sql/dev-fixture.sql',
            'share/sql/upgrade-monitoring-purge-instances.sql',
        ]),
        ('share/temboard/quickstart/', [
            'share/temboard_CHANGEME.key',
            'share/temboard_CHANGEME.pem',
            'share/temboard_ca_certs_CHANGEME.pem',
            'share/temboard.conf',
            'share/temboard.logrotate',
        ]),
        ('lib/systemd/system', ['packaging/temboard.service']),
    ],
    entry_points={
        'console_scripts': [
            'temboard = temboardui.__main__:main',
            'temboard-migratedb = temboardui.migratedb:main',
        ],
    },
)


if __name__ == '__main__':
    setup(
        long_description=open('README.rst').read(),
        packages=find_packages(),
        **SETUP_KWARGS
    )



import re
import sys
from platform import python_version


__version__ = "7.5"


VERSION_FMT = """\
temBoard %(temboard)s
System %(distname)s %(distversion)s
Python %(python)s (%(pythonbin)s)
psycopg2 %(psycopg2)s
Tornado %(tornado)s
SQLAlchemy %(sqlalchemy)s
alembic %(alembic)s
"""


def format_version():
    return VERSION_FMT % inspect_versions()


def inspect_versions():
    from alembic import __version__ as alembic_version
    from psycopg2 import __version__ as psycopg2_version
    from tornado import version as tornado_version
    from sqlalchemy import __version__ as sqlalchemy_version

    with open('/etc/os-release') as fo:
        distinfos = parse_lsb_release(fo)

    return dict(
        temboard=__version__,
        alembic=alembic_version,
        psycopg2=psycopg2_version,
        python=python_version(),
        pythonbin=sys.executable,
        tornado=tornado_version,
        sqlalchemy=sqlalchemy_version,
        distname=distinfos['NAME'],
        distversion=distinfos['VERSION'],
    )


def parse_lsb_release(lines):
    _assignement_re = re.compile(
        r"""(?P<variable>[A-Z_]+)="(?P<value>[^"]+)"$"""
    )
    infos = dict()
    for line in lines:
        m = _assignement_re.match(line)
        if not m:
            continue
        infos[m.group('variable')] = m.group('value')
    return infos



"""monitoring

Revision ID: 81f0aef8a119
Revises: e73c6bfce646
Create Date: 2020-03-25 16:50:03.723607

"""
import os.path

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '81f0aef8a119'
down_revision = 'e73c6bfce646'
branch_labels = None
depends_on = None


def sqlfile(name):
    directory = os.path.dirname(__file__)
    with open(os.path.join(directory, name + '.sql')) as fo:
        return sa.text(fo.read())


def upgrade():
    op.get_bind().execute(sqlfile('monitoring'))



import json
import logging
from os import path
import tornado.web
from tornado.escape import json_decode

from sqlalchemy.orm import (
    sessionmaker,
    scoped_session,
)
from sqlalchemy.sql import (
    column,
    extract,
    func,
    select,
    text,
)
from temboardui.model import worker_engine
from temboardui.model.orm import (
    Biggest,
    Biggestsum,
    Instances,
    diff,
    to_epoch,
    total_hit,
    total_read,
)
from temboardui.errors import TemboardUIError

from temboardui.web import (
    Blueprint,
    TemplateRenderer,
    jsonify,
)
from temboardui.plugins.monitoring.tools import (
    get_request_ids,
    parse_start_end,
)
from temboardui.toolkit import taskmanager
from temboardui.temboardclient import (
    temboard_request,
)


logger = logging.getLogger(__name__)
workers = taskmanager.WorkerSet()

PLUGIN_NAME = 'statements'

blueprint = Blueprint()
plugin_path = path.dirname(path.realpath(__file__))
render_template = TemplateRenderer(plugin_path + '/templates')


def configuration(config):
    return {}


def get_routes(config):
    routes = blueprint.rules + [
        (r"/js/statements/(.*)", tornado.web.StaticFileHandler, {
            'path': plugin_path + "/static/js"
        }),
    ]
    return routes


def get_agent_username(request):
    try:
        agent_username = request.instance.get_profile()['username']
    except Exception:
        agent_username = None
    return agent_username


METAS_QUERY = text("""
    SELECT *
    FROM statements.metas
    WHERE agent_address = :agent_address
    AND agent_port = :agent_port
""")


BASE_QUERY_STATDATA = text("""
    (
        SELECT dbid, datname, (record).*
        FROM (
          SELECT psh.dbid, psh.datname, unnest(records) AS record
          FROM statements.statements_history_db psh
          WHERE coalesce_range && tstzrange(:start, :end,'[]')
          AND    agent_address = :agent_address
          AND    agent_port = :agent_port
        ) AS unnested
        WHERE tstzrange((record).ts, (record).ts, '[]')
              <@ tstzrange(:start, :end, '[]')

        UNION ALL

        SELECT dbid, datname, (record).*
        FROM statements.statements_history_current_db
        WHERE agent_address = :agent_address
        AND agent_port = :agent_port
        AND tstzrange((record).ts, (record).ts, '[]')
          <@ tstzrange(:start, :end, '[]')
    ) h
""")


@blueprint.instance_route(r'/statements/data')
def json_data_instance(request):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)

    base_query = BASE_QUERY_STATDATA
    diffs = get_diffs_forstatdata()
    query = (select([
        column("datname"),
        column("dbid"),
    ] + diffs)
            .select_from(base_query)
            .group_by(column("dbid"), column("datname"))
            .having(func.max(column("calls")) - func.min(column("calls")) > 0))

    statements = request.db_session.execute(
        query,
        dict(agent_address=request.instance.agent_address,
             agent_port=request.instance.agent_port,
             start=start,
             end=end)) \
        .fetchall()
    statements = [dict(statement) for statement in statements]

    metas = request.db_session.execute(
        METAS_QUERY,
        dict(agent_address=request.instance.agent_address,
             agent_port=request.instance.agent_port)).fetchone()
    metas = dict(metas) if metas is not None else None
    return jsonify(dict(data=statements, metas=metas))


BASE_QUERY_STATDATA_DATABASE = """
WITH first_occurence AS (
  -- first occurence for each statement
  SELECT
    distinct on (queryid, userid)
    queryid,
    userid,
    ts,
    calls,
    total_exec_time,
    shared_blks_read,
    shared_blks_hit,
    shared_blks_dirtied,
    shared_blks_written,
    local_blks_read,
    local_blks_hit,
    local_blks_dirtied,
    local_blks_written,
    temp_blks_read,
    temp_blks_written,
    blk_read_time,
    blk_write_time
  FROM
    (
      (
        SELECT
          distinct on (queryid, userid)
          queryid,
          userid,
          (record).*
        FROM
          (
            SELECT
              distinct on (queryid, userid)
              queryid,
              userid,
              unnest(records) AS record
            FROM
              statements.statements_history psh
            WHERE
              coalesce_range && tstzrange(
                (:start) :: timestamptz,
                (:end) :: timestamptz,
                '[]'
              )
              AND agent_address = :agent_address
              AND agent_port = :agent_port
              AND dbid = :dbid
              {queryidfilter}
            order by
              queryid,
              userid,
              coalesce_range
          ) AS unnested
        WHERE
          tstzrange(
            (record).ts,
            (record).ts,
            '[]'
          ) <@ tstzrange(
            (:start) :: timestamptz,
            (:end) :: timestamptz,
            '[]'
          )
        order by
          queryid,
          userid,
          ts
      )
      UNION ALL
        (
          SELECT
            distinct on (queryid, userid)
            queryid,
            userid,
            (record).*
          FROM
            statements.statements_history_current
          WHERE
            agent_address = :agent_address
            AND agent_port = :agent_port
            AND dbid = :dbid
            AND tstzrange(
              (record).ts,
              (record).ts,
              '[]'
            ) <@ tstzrange(
              (:start) :: timestamptz,
              (:end) :: timestamptz,
              '[]'
            )
            {queryidfilter}
          order by
            queryid,
            userid,
            ts
        )
    ) as h
  order by
    queryid,
    userid,
    ts
),
last_occurence AS (
  -- last occurence for each statement
  SELECT
    distinct on (queryid, userid)
    queryid,
    userid,
    ts,
    calls,
    total_exec_time,
    shared_blks_read,
    shared_blks_hit,
    shared_blks_dirtied,
    shared_blks_written,
    local_blks_read,
    local_blks_hit,
    local_blks_dirtied,
    local_blks_written,
    temp_blks_read,
    temp_blks_written,
    blk_read_time,
    blk_write_time
  FROM
    (
      (
        SELECT
          distinct on (queryid, userid)
          queryid,
          userid,
          (record).*
        FROM
          (
            SELECT
              distinct on (queryid, userid)
              queryid,
              userid,
              unnest(records) AS record
            FROM
              statements.statements_history psh
            WHERE
              coalesce_range && tstzrange(
                (:start) :: timestamptz,
                (:end) :: timestamptz,
                '[]'
              )
              AND agent_address = :agent_address
              AND agent_port = :agent_port
              AND dbid = :dbid
              {queryidfilter}
            order by
              queryid,
              userid,
              coalesce_range desc
          ) AS unnested
        WHERE
          tstzrange(
            (record).ts,
            (record).ts,
            '[]'
          ) <@ tstzrange(
            (:start) :: timestamptz,
            (:end) :: timestamptz,
            '[]'
          )
        order by
          queryid,
          userid,
          ts desc
      )
      UNION ALL
        (
          SELECT
            distinct on (queryid, userid)
            queryid,
            userid,
            (record).*
          FROM
            statements.statements_history_current
          WHERE
            agent_address = :agent_address
            AND agent_port = :agent_port
            AND dbid = :dbid
            AND tstzrange(
              (record).ts,
              (record).ts,
              '[]'
            ) <@ tstzrange(
              (:start) :: timestamptz,
              (:end) :: timestamptz,
              '[]'
            )
            {queryidfilter}
          order by
            queryid,
            userid,
            ts desc
        )
    ) as h
  order by
    queryid,
    userid,
    ts desc
)
SELECT
  query,
  statements.queryid::text,
  rolname,
  statements.userid::text,
  (lo.calls - fo.calls) AS calls,
  (lo.total_exec_time - fo.total_exec_time) AS total_exec_time,
  (lo.total_exec_time - fo.total_exec_time) /
    (lo.calls - fo.calls) AS mean_time,
  (
    lo.shared_blks_read - fo.shared_blks_read
  ) AS shared_blks_read,
  (
    lo.shared_blks_hit - fo.shared_blks_hit
  ) AS shared_blks_hit,
  (
    lo.shared_blks_dirtied - fo.shared_blks_dirtied
  ) AS shared_blks_dirtied,
  (
    lo.shared_blks_written - fo.shared_blks_written
  ) AS shared_blks_written,
  (
    lo.local_blks_read - fo.local_blks_read
  ) AS local_blks_read,
  (
    lo.local_blks_hit - fo.local_blks_hit
  ) AS local_blks_hit,
  (
    lo.local_blks_dirtied - fo.local_blks_dirtied
  ) AS local_blks_dirtied,
  (
    lo.local_blks_written - fo.local_blks_written
  ) AS local_blks_written,
  (
    lo.temp_blks_read - fo.temp_blks_read
  ) AS temp_blks_read,
  (
    lo.temp_blks_written - fo.temp_blks_written
  ) AS temp_blks_written,
  (
    lo.blk_read_time - fo.blk_read_time
  ) AS blk_read_time,
  (
    lo.blk_write_time - fo.blk_write_time
  ) AS blk_write_time
FROM
  first_occurence AS fo
  JOIN last_occurence AS lo
    ON fo.queryid = lo.queryid
    AND fo.userid = lo.userid
  JOIN statements.statements
    ON statements.queryid = fo.queryid
    AND statements.userid = fo.userid
  AND agent_address = :agent_address
  AND agent_port = :agent_port
  AND dbid = :dbid
WHERE (lo.calls - fo.calls) > 0;
"""


@blueprint.instance_route(r'/statements/data/([0-9]*)/([-]?[0-9]*)/([0-9]*)')
def json_data_query(request, dbid, queryid, userid):
    return json_data(request, dbid, queryid, userid)


@blueprint.instance_route(r'/statements/data/(.*)')
def json_data_database(request, dbid):
    return json_data(request, dbid)


def json_data(request, dbid, queryid=None, userid=None):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)

    query = text("""
        SELECT DISTINCT(datname)
        FROM statements.statements
        WHERE dbid = :dbid
        AND agent_address = :agent_address
        AND agent_port = :agent_port;
    """)
    datname = request.db_session.execute(
        query,
        dict(agent_address=request.instance.agent_address,
             agent_port=request.instance.agent_port,
             dbid=dbid)
    ).fetchone()[0]

    params = dict(agent_address=request.instance.agent_address,
                  agent_port=request.instance.agent_port,
                  dbid=dbid,
                  start=start,
                  end=end)

    query = BASE_QUERY_STATDATA_DATABASE
    queryidfilter = ''
    if queryid is not None and userid is not None:
        queryidfilter = 'AND queryid = :queryid AND userid = :userid'
        params.update(dict(queryid=queryid, userid=userid))
    query = query.format(**dict(queryidfilter=queryidfilter))

    statements = request.db_session.execute(query, params).fetchall()
    statements = [dict(statement) for statement in statements]
    return jsonify(dict(datname=datname, data=statements))


def get_diffs_forstatdata():
    return [
        diff("calls"),
        diff("total_exec_time").label("total_exec_time"),
        (diff("total_exec_time") / diff("calls")).label("mean_time"),
        diff("shared_blks_read"),
        diff("shared_blks_hit"),
        diff("shared_blks_dirtied"),
        diff("shared_blks_written"),
        diff("local_blks_read"),
        diff("local_blks_hit"),
        diff("local_blks_dirtied"),
        diff("local_blks_written"),
        diff("temp_blks_read"),
        diff("temp_blks_written"),
        diff("blk_read_time"),
        diff("blk_write_time")
    ]


BASE_QUERY_STATDATA_SAMPLE_INSTANCE = text("""
    (
      SELECT *
      FROM (
        SELECT
          row_number() OVER (ORDER BY ts) AS number,
          count(*) OVER (PARTITION BY 1) AS total,
          *
        FROM (
          SELECT
            ts,
            sum(calls) AS calls,
            sum(total_exec_time) AS total_exec_time,
            sum(rows) AS rows,
            sum(shared_blks_hit) AS shared_blks_hit,
            sum(shared_blks_read) AS shared_blks_read,
            sum(shared_blks_dirtied) AS shared_blks_dirtied,
            sum(shared_blks_written) AS shared_blks_written,
            sum(local_blks_hit) AS local_blks_hit,
            sum(local_blks_read) AS local_blks_read,
            sum(local_blks_dirtied) AS local_blks_dirtied,
            sum(local_blks_written) AS local_blks_written,
            sum(temp_blks_read) AS temp_blks_read,
            sum(temp_blks_written) AS temp_blks_written,
            sum(blk_read_time) AS blk_read_time,
            sum(blk_write_time) AS blk_write_time
          FROM (
            SELECT (record).*
            FROM (
              SELECT psh.dbid, psh.coalesce_range, unnest(records) AS record
              FROM statements.statements_history_db psh
              WHERE coalesce_range && tstzrange(:start, :end,'[]')
              AND    agent_address = :agent_address
              AND    agent_port = :agent_port
            ) AS unnested
            WHERE tstzrange((record).ts, (record).ts, '[]')
                  <@ tstzrange(:start, :end, '[]')

            UNION ALL

            SELECT (record).*
            FROM statements.statements_history_current_db
            WHERE tstzrange((record).ts, (record).ts, '[]')
              <@ tstzrange(:start, :end, '[]')
            AND agent_address = :agent_address
            AND agent_port = :agent_port
          ) AS s
          GROUP BY ts
        ) AS statements_history
      ) AS sh
      WHERE number % ( int8larger((total)/(:samples +1),1) ) = 0
    ) by_instance
""")


BASE_QUERY_STATDATA_SAMPLE_DATABASE = text("""
    (
      SELECT *
      FROM (
        SELECT
          row_number() OVER (ORDER BY ts) AS number,
          count(*) OVER (PARTITION BY 1) AS total,
          *
        FROM (
          SELECT (record).*
          FROM (
            SELECT psh.dbid, psh.coalesce_range, unnest(records) AS record
            FROM statements.statements_history_db psh
            WHERE coalesce_range && tstzrange(:start, :end,'[]')
            AND    dbid = :dbid
            AND    agent_address = :agent_address
            AND    agent_port = :agent_port
          ) AS unnested
          WHERE tstzrange((record).ts, (record).ts, '[]')
                <@ tstzrange(:start, :end, '[]')

          UNION ALL

          SELECT (record).*
          FROM statements.statements_history_current_db
          WHERE tstzrange((record).ts, (record).ts, '[]')
            <@ tstzrange(:start, :end, '[]')
          AND dbid = :dbid
          AND agent_address = :agent_address
          AND agent_port = :agent_port
        ) AS statements_history
      ) AS sh
      WHERE number % ( int8larger((total)/(:samples +1),1) ) = 0
    ) by_db
""")

BASE_QUERY_STATDATA_SAMPLE_QUERY = text("""
    (
      SELECT *
      FROM (
        SELECT
          row_number() OVER (ORDER BY ts) AS number,
          count(*) OVER (PARTITION BY 1) AS total,
          *
        FROM (
          SELECT (record).*
          FROM (
            SELECT psh.dbid, psh.coalesce_range, unnest(records) AS record
            FROM statements.statements_history psh
            WHERE coalesce_range && tstzrange(:start, :end,'[]')
            AND dbid = :dbid
            AND queryid = :queryid
            AND userid = :userid
            AND agent_address = :agent_address
            AND agent_port = :agent_port
          ) AS unnested
          WHERE tstzrange((record).ts, (record).ts, '[]')
                <@ tstzrange(:start, :end, '[]')

          UNION ALL

          SELECT (record).*
          FROM statements.statements_history_current
          WHERE tstzrange((record).ts, (record).ts, '[]')
            <@ tstzrange(:start, :end, '[]')
          AND dbid = :dbid
          AND queryid = :queryid
          AND userid = :userid
          AND agent_address = :agent_address
          AND agent_port = :agent_port
        ) AS statements_history
      ) AS sh
      WHERE number % ( int8larger((total)/(:samples +1),1) ) = 0
    ) by_db
""")


def getstatdata_sample(request, mode, start, end, dbid=None, queryid=None,
                       userid=None):
    if mode == 'instance':
        base_query = BASE_QUERY_STATDATA_SAMPLE_INSTANCE

    elif mode == "db":
        base_query = BASE_QUERY_STATDATA_SAMPLE_DATABASE

    elif mode == "query":
        base_query = BASE_QUERY_STATDATA_SAMPLE_QUERY

    ts = column('ts')
    biggest = Biggest(ts)
    biggestsum = Biggestsum(ts)

    subquery = (select([
        ts,
        biggest("ts", '0 s', "mesure_interval"),
        biggestsum("calls"),
        biggestsum("total_exec_time", label="runtime"),
        biggestsum("rows"),
        biggestsum("shared_blks_read"),
        biggestsum("shared_blks_hit"),
        biggestsum("shared_blks_dirtied"),
        biggestsum("shared_blks_written"),
        biggestsum("local_blks_read"),
        biggestsum("local_blks_hit"),
        biggestsum("local_blks_dirtied"),
        biggestsum("local_blks_written"),
        biggestsum("temp_blks_read"),
        biggestsum("temp_blks_written"),
        biggestsum("blk_read_time"),
        biggestsum("blk_write_time")
        ])
            .select_from(base_query)
            .apply_labels()
            .group_by(*([ts])))

    subquery = subquery.alias()
    c = subquery.c

    greatest = func.greatest
    cols = [
        to_epoch(c.ts),
        (
            func.sum(c.calls) /
            greatest(extract("epoch", c.mesure_interval), 1)
        ).label("calls"),
        (
            func.sum(c.runtime) / greatest(func.sum(c.calls), 1.)
        ).label("avg_runtime"),
        (
            func.sum(c.runtime) /
            greatest(extract("epoch", c.mesure_interval), 1)
        ).label("load"),
        total_read(c),
        total_hit(c)
    ]

    query = (
        select(cols)
        .select_from(subquery)
        .where(c.calls != 0)
        .group_by(c.ts, c.mesure_interval)
        .order_by(c.ts)
    )

    params = dict(agent_address=request.instance.agent_address,
                  agent_port=request.instance.agent_port,
                  samples=50,
                  start=start,
                  end=end)

    if mode == 'db' or mode == 'query':
        params['dbid'] = dbid
    if mode == 'query':
        params['queryid'] = queryid
        params['userid'] = userid

    rows = request.db_session.execute(query, params).fetchall()
    return [dict(row) for row in rows]


@blueprint.instance_route(r'/statements/chart')
def json_chart_data_instance(request):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)

    data = getstatdata_sample(request, "instance", start, end)
    return jsonify(dict(data=data))


@blueprint.instance_route(r'/statements/chart/([0-9]*)/([-]?[0-9]*)/([0-9]*)')
def json_chart_data_query(request, dbid, queryid, userid):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)

    data = getstatdata_sample(request, "query", start, end, dbid=dbid,
                              queryid=queryid, userid=userid)
    return jsonify(dict(data=data))


@blueprint.instance_route(r'/statements/chart/(.*)')
def json_chart_data_db(request, dbid):
    host_id, instance_id = get_request_ids(request)
    start, end = parse_start_end(request)

    data = getstatdata_sample(request, "db", start, end, dbid=dbid)
    return jsonify(dict(data=data))


@blueprint.instance_route(r'/statements')
def statements(request):
    request.instance.check_active_plugin(PLUGIN_NAME)
    return render_template(
        'index.html',
        nav=True,
        agent_username=get_agent_username(request),
        instance=request.instance,
        plugin=__name__,
        role=request.current_user,
    )


def add_statement(session, instance, data):
    try:
        cur = session.connection().connection.cursor()
        cur.execute("SET search_path TO statements")
        for statement in data.get('data'):
            query = """
                INSERT INTO statements_src_tmp
                VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                        %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
            """
            cur.execute(
                query,
                (
                    instance.agent_address,
                    instance.agent_port,
                    data['snapshot_datetime'],
                    statement['userid'],
                    statement['rolname'],
                    statement['dbid'],
                    statement['datname'],
                    statement['queryid'],
                    statement['query'],
                    statement['calls'],
                    statement['total_exec_time']
                    if 'total_exec_time' in statement
                    else statement['total_time'],
                    statement['rows'],
                    statement['shared_blks_hit'],
                    statement['shared_blks_read'],
                    statement['shared_blks_dirtied'],
                    statement['shared_blks_written'],
                    statement['local_blks_hit'],
                    statement['local_blks_read'],
                    statement['local_blks_dirtied'],
                    statement['local_blks_written'],
                    statement['temp_blks_read'],
                    statement['temp_blks_written'],
                    statement['blk_read_time'],
                    statement['blk_write_time'],
                    statement['total_plan_time']
                    if 'total_plan_time' in statement
                    else None,
                    statement['wal_records']
                    if 'wal_records' in statement
                    else None,
                    statement['wal_fpi']
                    if 'wal_fpi' in statement
                    else None,
                    statement['wal_bytes']
                    if 'wal_bytes' in statement
                    else None,
                )
            )
        query = """SELECT process_statements(%s, %s)"""
        cur.execute(query, (instance.agent_address, instance.agent_port))
        session.connection().connection.commit()
    except Exception as e:
        raise TemboardUIError(400, e.message)


@workers.register(pool_size=1)
def pull_data_worker(app):
    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    worker_session = Session()
    instances = worker_session.query(Instances)

    for instance in instances:
        plugin_names = [plugin.plugin_name for plugin in instance.plugins]

        if 'statements' not in plugin_names:
            continue

        # FIXME should be done in parallel
        try:
            pull_data_for_instance(app, worker_session, instance)
        except Exception:
            pass


def pull_data_for_instance(app, session, instance):
    url = 'https://%s:%s%s?key=%s' % (
        instance.agent_address,
        instance.agent_port,
        '/statements',
        instance.agent_key,
    )
    headers = {}
    try:
        body = temboard_request(
            app.config.temboard.ssl_ca_cert_file,
            method='GET',
            url=url,
            headers=headers,
        )
        add_statement(session, instance, json_decode(body))
    except Exception as e:
        error = 'Error while fetching statements from instance: '
        if hasattr(e, 'read'):
            error += json.loads(e.read())['error']
        else:
            error += str(e)

        logger.error(error)
        logger.exception(e)

        # If statements data cannot be retrieved store the error in the
        # statements metas table
        cur = session.connection().connection.cursor()
        cur.execute("SET search_path TO statements")
        query = """
            -- Create new meta for agent if doesn't already exist
            INSERT INTO metas (agent_address, agent_port)
            VALUES (%s, %s)
            ON CONFLICT DO NOTHING;
        """
        cur.execute(
            query,
            (
                instance.agent_address,
                instance.agent_port,
            )
        )

        query = """
            UPDATE metas
            SET error = %s
            WHERE agent_address = %s AND agent_port = %s;
        """
        cur.execute(
            query,
            (
                error,
                instance.agent_address,
                instance.agent_port,
            )
        )
        session.connection().connection.commit()
        raise(e)


@workers.register(pool_size=1)
def purge_data_worker(app):
    """Background worker in charge of purging statements data.
    Purge policy is based on purge_after parameter from statements section.
    purge_after sets the number of days of data to keep, from now. Default is
    7 days if not set.
    """
    logger.setLevel(app.config.logging.level)
    logger.info("Purging old statements data")

    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    session = Session()

    purge_after = app.config.statements.purge_after

    # Get tablename list to purge from metric_tables_config()
    try:
        cur = session.connection().connection.cursor()
        cur.execute("SET search_path TO statements")
        cur.execute("""SELECT statements_purge(%s)""", (purge_after,))
        session.connection().connection.commit()
    except Exception as e:
        logger.error('Could not purge statements data')
        logger.exception(e)
        raise(e)


@taskmanager.bootstrap()
def statements_bootstrap(context):
    yield taskmanager.Task(
        worker_name='pull_data_worker',
        id='statementsdata',
        redo_interval=1 * 60,  # Repeat each 1m,
        options={},
    )
    yield taskmanager.Task(
        worker_name='purge_data_worker',
        id='purge_data',
        redo_interval=24 * 60 * 60,  # Repeat each 24h,
        options={},
    )



import operator


def bootstrap_checks(hostinfo):
    # Default checks with thresholds to run against monitoring data

    # Loadaverage (value)
    # Global to host
    yield (u"load1", hostinfo['n_cpu'] / float(2), hostinfo['n_cpu'])
    # CPU (percent)
    # One per CPU
    yield (u"cpu_core", 50, 80)
    # Memory usage (percent)
    # global
    yield (u"memory_usage", 50, 80)
    # Swap usage (percent)
    # Global to host
    yield (u"swap_usage", 30, 50)
    # Filesystems usage (percent)
    # One per filesystem
    yield (u"fs_usage_mountpoint", 80, 90)
    # Number of WAL files ready to be archived (value)
    # Global to postgres instance
    yield (u"wal_files_archive", 10, 20)
    # Number of WAL files (value)
    # Global to postgres instance
    yield (u"wal_files_total", 50, 100)
    # Number of transaction rollback (value)
    # One per DB
    yield (u"rollback_db", 10, 20)
    # Cache hitratio (percent)
    # One per DB
    yield (u"hitreadratio_db", 90, 80)
    # Client sessions vs max_connections (percent)
    # Global to postgres instance
    yield (u"sessions_usage", 80, 90)
    # Waiting sessions (value)
    # One per DB
    yield (u"waiting_sessions_db", 5, 10)
    # Replication lag
    # warning: 1MB, critical: 10MB
    yield (u"replication_lag", 1024 * 1024, 10 * 1024 * 1024)
    # Replication connection
    yield (u"replication_connection", 0, 0)
    # Temporary files size delta
    # warning: 1MB, critical: 10MB
    yield (u"temp_files_size_delta", 1024 * 1024, 10 * 1024 * 1024)
    # Heap bloat ratio
    yield (u"heap_bloat", 30, 50)
    # Btree bloat ratio
    yield (u"btree_bloat", 30, 50)


class PreProcess(object):

    @staticmethod
    def loadaverage(data):
        return float(data['loadavg'][0]['load1'])

    @staticmethod
    def cpu(data):
        _data = dict()
        for r in data['cpu']:
            total = 0
            idle = 0
            total += int(r['time_system'])
            total += int(r['time_steal'])
            total += int(r['time_iowait'])
            total += int(r['time_user'])
            total += int(r['time_idle'])
            idle = int(r['time_idle'])
            _data[r['cpu']] = int((total - idle) / float(total) * 100)
        return _data

    @staticmethod
    def memory(data):
        return int(
            (int(data['memory'][0]['mem_total'])
             - int(data['memory'][0]['mem_free'])
             - int(data['memory'][0]['mem_cached']))
            / float(data['memory'][0]['mem_total']) * 100
        )

    @staticmethod
    def swap(data):
        return int(
            int(data['memory'][0]['swap_used'])
            / float(data['memory'][0]['swap_total']) * 100
        )

    @staticmethod
    def fs(data):
        _data = dict()
        for r in data['filesystems_size']:
            _data[r['mount_point']] = int(int(r['used']) / float(r['total'])
                                          * 100)
        return _data

    @staticmethod
    def archive_ready(data):
        return int(data['wal_files'][0]['archive_ready'])

    @staticmethod
    def wal_files(data):
        return int(data['wal_files'][0]['total'])

    @staticmethod
    def xacts_rollback(data):
        _data = dict()
        for r in data['xacts']:
            _data[r['dbname']] = int(r['n_rollback'])
        return _data

    @staticmethod
    def hitratio(data):
        _data = dict()
        for r in data['blocks']:
            hit = int(r['blks_hit'])
            read = int(r['blks_read'])
            _data[r['dbname']] = (100 * hit / (hit + read)) \
                if read + hit > 0 else 100
        return _data

    @staticmethod
    def sessions(data):
        n = 0
        for r in data['sessions']:
            n += int(r['idle_in_xact'])
            n += int(r['idle_in_xact_aborted'])
            n += int(r['no_priv'])
            n += int(r['idle'])
            n += int(r['disabled'])
            n += int(r['waiting'])
            n += int(r['active'])
            n += int(r['fastpath'])
        return int(n / float(data['max_connections']) * 100)

    @staticmethod
    def waiting(data):
        _data = dict()
        for r in data['sessions']:
            _data[r['dbname']] = int(r['waiting'])
        return _data

    @staticmethod
    def replication_lag(data):
        return int(data['replication_lag'][0]['lag'])

    @staticmethod
    def replication_connection(data):
        k = data['replication_connection'][0]['upstream']
        v = int(data['replication_connection'][0]['connected'])
        return {k: v}

    @staticmethod
    def temp_files_size_delta(data):
        _data = dict()
        for r in data['temp_files_size_delta']:
            _data[r['dbname']] = int(r['size'])
        return _data

    @staticmethod
    def heap_bloat(data):
        _data = dict()
        for r in data['heap_bloat']:
            _data[r['dbname']] = int(r['ratio'])
        return _data

    @staticmethod
    def btree_bloat(data):
        _data = dict()
        for r in data['btree_bloat']:
            _data[r['dbname']] = int(r['ratio'])
        return _data


check_specs = dict(
    load1=dict(
        category='system',
        description='Loadaverage',
        preprocess=PreProcess.loadaverage,
        message='{value} is greater than {threshold}',
        operator=operator.gt,
    ),
    cpu_core=dict(
        category='system',
        description='CPU usage',
        preprocess=PreProcess.cpu,
        message='{value}% is greater than {threshold}%',
        operator=operator.gt,
        value_type='percent',
    ),
    memory_usage=dict(
        category='system',
        description='Memory usage',
        preprocess=PreProcess.memory,
        message='{value}% is greater than {threshold}%',
        operator=operator.gt,
        value_type='percent'
    ),
    swap_usage=dict(
        category='system',
        description='Swap usage',
        preprocess=PreProcess.swap,
        message='{value}% is greater than {threshold}%',
        operator=operator.gt,
        value_type='percent',
    ),
    fs_usage_mountpoint=dict(
        category='system',
        description='File systems usage',
        preprocess=PreProcess.fs,
        message='{key}: {value}% is greater than {threshold}%',
        operator=operator.gt,
        value_type='percent',
    ),
    wal_files_archive=dict(
        category='postgres',
        description='WAL files ready to be archived',
        preprocess=PreProcess.archive_ready,
        message='{value} is greater than {threshold}',
        operator=operator.gt,
    ),
    wal_files_total=dict(
        category='postgres',
        description='WAL files',
        preprocess=PreProcess.wal_files,
        message='{value} is greater than {threshold}',
        operator=operator.gt,
    ),
    rollback_db=dict(
        category='postgres',
        description='Rollbacked transactions',
        preprocess=PreProcess.xacts_rollback,
        message='{key}: {value} is greater than {threshold}',
        operator=operator.gt,
    ),
    hitreadratio_db=dict(
        category='postgres',
        description='Cache Hit Ratio',
        preprocess=PreProcess.hitratio,
        message='{key}: {value} is less than {threshold}',
        operator=operator.lt,
        value_type='percent',
    ),
    sessions_usage=dict(
        category='postgres',
        description='Client sessions',
        preprocess=PreProcess.sessions,
        message='{value} is greater than {threshold}',
        operator=operator.gt,
        value_type='percent',
    ),
    waiting_sessions_db=dict(
        category='postgres',
        description='Waiting sessions',
        preprocess=PreProcess.waiting,
        message='{key}: {value} is greater than {threshold}',
        operator=operator.gt,
    ),
    replication_lag=dict(
        category='postgres',
        description='Streaming replication lag',
        preprocess=PreProcess.replication_lag,
        message='{key}: {value} is greater than {threshold}',
        operator=operator.gt,
    ),
    replication_connection=dict(
        category='postgres',
        description='Streaming replication connection',
        preprocess=PreProcess.replication_connection,
        message='{key}: not connected',
        operator=operator.eq,
    ),
    temp_files_size_delta=dict(
        category='postgres',
        description='Temporary files size (delta)',
        preprocess=PreProcess.temp_files_size_delta,
        message='{key}: {value} is greater than {threshold}',
        operator=operator.gt,
    ),
    heap_bloat=dict(
        category='postgres',
        description='Heap bloat ratio estimation',
        preprocess=PreProcess.heap_bloat,
        message='{key}: {value} is greater than {threshold}',
        operator=operator.gt,
    ),
    btree_bloat=dict(
        category='postgres',
        description='BTree index bloat ratio estimation',
        preprocess=PreProcess.btree_bloat,
        message='{key}: {value} is greater than {threshold}',
        operator=operator.gt,
    ),
)


def get_highest_state(states):
    """
    Returns the highest state.
    """
    levels = ['UNDEF', 'OK', 'WARNING', 'CRITICAL']
    return levels[max([levels.index(state) for state in states])]


def checks_info(session, host_id, instance_id):
    """
    Returns alerting checks with current state by host_id/instance_id
    """
    query = """
SELECT c.name, c.warning, c.critical, c.description, c.enabled,
json_agg(json_build_object('key', cs.key, 'state', cs.state)) AS state_by_key
FROM monitoring.checks c JOIN monitoring.check_states cs ON (c.check_id = cs.check_id)
WHERE host_id = :host_id AND instance_id = :instance_id
GROUP BY 1,2,3,4,5 ORDER BY 1
    """  # noqa
    res = session.execute(query,
                          dict(host_id=host_id, instance_id=instance_id))
    ret = []
    for row in res.fetchall():
        c_row = dict(row)
        c_row['state'] = get_highest_state([o['state']
                                            for o in c_row['state_by_key']])
        ret.append(c_row)
    return ret


def check_state_detail(session, host_id, instance_id, check_name):
    query = """
SELECT json_agg(json_build_object('key', cs.key,
                                  'state', cs.state,
                                  'enabled', c.enabled,
                                  'datetime', sc.datetime,
                                  'value', sc.value,
                                  'warning', sc.warning,
                                  'critical', sc.critical)) AS state_detail
FROM monitoring.checks c JOIN monitoring.check_states cs ON (c.check_id = cs.check_id)
JOIN monitoring.state_changes sc ON (sc.check_id = c.check_id AND sc.key = cs.key
                                     AND sc.datetime = (SELECT MAX(datetime)
                                                        FROM monitoring.state_changes sc2
                                                        WHERE sc2.check_id=c.check_id
                                                        AND sc2.key = cs.key
                                                        AND sc2.state = cs.state))
WHERE host_id = :host_id AND instance_id = :instance_id AND c.name = :check_name
    """  # noqa
    res = session.execute(query,
                          dict(host_id=host_id, instance_id=instance_id,
                               check_name=check_name))
    row = res.fetchone()
    c_row = dict(row)
    return c_row['state_detail']
